//=============================================================================
// ActorInfo.
// 
// Date: 2016
// ©copyright Francesco Biscazzo. All rights reserved.
//=============================================================================
class ActorInfo extends TournamentPickup;

var ValueRegister ValueRegister; //ValueRegister chain.

var AdvancedMutatorManager AMM; //The AdvancedMutator this AI is registered to.
var ActorInfoPickup AIP; //The ActorInfoPickup this AI spawned.
var int noneTimes; //The ticks amount the owner of this AI has been None.
var bool bNoneNotified; //True if this AI called BecameNone() and the owner is still None.

var String ownerState;
var float spawnedTime;
var Weapon ownerWeapon;
var Actor oldAimedActor;
var vector oldHitLocation, oldHitNormal, oldStartTrace, oldEndTrace;
var ZoneInfo ownerZone;
var TournamentPickup lastAffector; //Last picked up Affector.

var bool bChagingWeapon;

function Spawned() {
	super.Spawned();
	
	spawnedTime = Level.timeSeconds;
}
function Destroyed() {
	local ValueRegister nextValueRegister;

	AMM.BaseAM.AIDestroying(Self, Owner);
	
	if (AIP != None)
		AIP.Destroy();
	
	//Destroy all the ValueRegister of this AI.
	for(nextValueRegister = ValueRegister; nextValueRegister != None; nextValueRegister = nextValueRegister.ValueRegister)
		nextValueRegister.Destroy();
	
	//Put the last affector back in the Affector variable of the owner weapon(if it's a TournamentWeapon).
	if (lastAffector != None)
		if (ownerWeapon != None)
			if (ownerWeapon.IsA('TournamentWeapon'))
				TournamentWeapon(ownerWeapon).Affector = lastAffector;
	
	super.Destroyed();
}

//Give an AIP to the Owner.
function GiveAIP() {
	local Actor nextAct;
	local bool bIsInChain;
	
	if ((Owner != None) && (!Owner.IsA('Inventory'))) {
		if (AIP == None)
			AIP = Spawn(class'ActorInfoPickup', Owner);
		if (AIP != None) {
			AIP.bHeldItem = true;
			AIP.RegisterActorInfo(self, AMM);
				
			if (Owner.IsA('Pawn'))
				AIP.Instigator = Pawn(Owner);
			
			AIP.Activate();
		}
	}
}

//Register this AI to an AdvancedMutator.
function RegisterAdvancedMutator(AdvancedMutatorManager AdvMutManager) {
	AMM = AdvMutManager;
	
	AMM.BaseAM.AIRegistered(Self, Owner);
	
	//Call PostFire if this AI's owner is a Projectile and it was registered at the time it was spawned.
	if ((Level.timeSeconds - spawnedTime) < 1)
		if (Owner.IsA('Projectile'))
			if (Owner.Instigator != None)
					AMM.BaseAM.PostFire(Owner.Instigator, Owner.Instigator.Weapon, AMM.BaseAM.static.getAimedActor(Owner.Instigator, Owner.Instigator.Weapon), Projectile(Owner));
}

function bool registerTextValue(class<AdvancedMutator> requesterClass, String valueName, String value, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	local ValueRegister vRegister, tempVRegister;
	
	//Check if already registered.
	getRegisteredValue(requesterClass, valueName, channel, tempVRegister);
	if (tempVRegister == None) {
		if (valueName != "") {
			vRegister = Spawn(class'ValueRegister');
			if (vRegister != None) {
				vRegister.bCheckStatus = bCheckStatus;
				vRegister.bUnchangedLoop = bUnchangedLoop;
				if (vRegister.registerTextValue(requesterClass, valueName, value, channel)) {
					//Add to ValueRegister chain.
					vRegister.ValueRegister = ValueRegister;
					ValueRegister = vRegister;
					
					return true;
				}
			}
		}
	} else {
		//A ValueRegister has this property value registered already and it was requested by the same class requesting this saving, update its value.
		tempVRegister.registerTextValue(requesterClass, valueName, value, channel);
	}
	
	return false;
}

//register the current value of the specified property of the Owner.
//For the use of bCheckStatus and bUnchangedLoop see CheckValues().
//channel is needed to be able to register multiple values for one property only.
function bool registerOwnerValue(class<AdvancedMutator> requesterClass, String propertyName, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	local ValueRegister vRegister, tempVRegister;
	
	//Check if already registered.
	getRegisteredValue(requesterClass, propertyName, channel, tempVRegister);
	if (tempVRegister == None) {
		if ((Owner != None) && (propertyName != "")) {
			vRegister = Spawn(class'ValueRegister');
			if (vRegister != None) {
				vRegister.bCheckStatus = bCheckStatus;
				vRegister.bUnchangedLoop = bUnchangedLoop;
				if (vRegister.registerVariableValue(requesterClass, Owner, propertyName, channel)) {
					//Add to ValueRegister chain.
					vRegister.ValueRegister = ValueRegister;
					ValueRegister = vRegister;
					
					return true;
				}
			}
		}
	} else {
		//A ValueRegister has this property value registered already and it was requested by the same class requesting this saving, update its value.
		tempVRegister.registerVariableValue(requesterClass, Owner, propertyName, channel);
	}
	
	return false;
}

//Get a previously registered value of the specified property of the Owner or just having the specified "valueName".
function String getRegisteredValue(class<AdvancedMutator> requesterClass, String propertyName, optional int channel, optional out ValueRegister vRegister) {
	local ValueRegister nextValueRegister;
	for(nextValueRegister = ValueRegister; nextValueRegister != None; nextValueRegister = nextValueRegister.ValueRegister) {
		if ((nextValueRegister.propertyName ~= propertyName) && bool(AMM.BaseAM.static.ClassIsRelationedWith(nextValueRegister.requesterClass, requesterClass)) && (nextValueRegister.channel == channel)) {
			vRegister = nextValueRegister;
			
			return nextValueRegister.propertyValue;
		}
	}
			
	return "";
}

//Remove a previously registered value of the specified property of the Owner or just having the specified "valueName".
function bool removeRegisteredValue(class<AdvancedMutator> requesterClass, String propertyName, optional int channel) {
	local ValueRegister prevValueRegister, nextValueRegister;
	local ValueRegister vRegister;
	
	if (ValueRegister == None)
		return false;
	
	if ((ValueRegister.propertyName ~= propertyName) && bool(AMM.BaseAM.static.ClassIsRelationedWith(ValueRegister.requesterClass, requesterClass)) && (ValueRegister.channel == channel)) {
		vRegister = ValueRegister.ValueRegister;
		ValueRegister.Destroy();
		ValueRegister = vRegister;
		
		return true;
	} else {
		prevValueRegister = ValueRegister;
		for(nextValueRegister = ValueRegister.ValueRegister; nextValueRegister != None; nextValueRegister = nextValueRegister.ValueRegister) {
			if ((nextValueRegister.propertyName ~= propertyName) && bool(AMM.BaseAM.static.ClassIsRelationedWith(nextValueRegister.requesterClass, requesterClass)) && (nextValueRegister.channel == channel)) {
				//Remove from ValueRegister chain.
				vRegister = nextValueRegister.ValueRegister;
				nextValueRegister.Destroy();
				prevValueRegister.ValueRegister = vRegister;
						
				return true;
			}
			
			prevValueRegister = nextValueRegister;
		}
	}
			
	return false;
}

//Check if a registered Owner value has been changed or not. If so, update the value within the ValueRegister.
//If bCheckStatus is true, the value will be updated every tick and changes will be notified through the calling of ValueStatusChanged(),
//if, in one tick, there haven't been any changes, it will be notified(once only if bUnchangedLoop is false) through the calling of ValueStatusUnchanged().
function checkValues() {
	local ValueRegister nextValueRegister;
	local String oldValue, newValue;
	
	for (nextValueRegister = ValueRegister; nextValueRegister != None; nextValueRegister = nextValueRegister.ValueRegister)
		if (nextValueRegister.bRealValue) {
			//If the value comes from an existing actor's property then check it.
			if (nextValueRegister.bCheckStatus) {
				oldValue = nextValueRegister.propertyValue;
				if (Owner != None)
					newValue = Owner.getPropertyText(nextValueRegister.propertyName);
				else
					newValue = "None";
				
				if (oldValue != newValue) {
					//Update value.
					registerOwnerValue(nextValueRegister.requesterClass, nextValueRegister.propertyName, nextValueRegister.bCheckStatus, nextValueRegister.bUnchangedLoop, nextValueRegister.channel);
					
					AMM.BroadcastValueStatusChanged(nextValueRegister.requesterClass, Owner, nextValueRegister.propertyName, oldValue, newValue, nextValueRegister.channel);
					nextValueRegister.bUnchanged = false;
				} else {
					if (!nextValueRegister.bUnchanged || (nextValueRegister.bUnchanged && nextValueRegister.bUnchangedLoop)) {
						AMM.BroadcastValueStatusUnchanged(nextValueRegister.requesterClass, Owner, nextValueRegister.propertyName, newValue, nextValueRegister.lastChangeTime, nextValueRegister.channel);
						nextValueRegister.bUnchanged = true;
					}
				}
			}
		}
}

function Activate() {
	GotoState('Activated');
	Activate();
}

function Deactivate() {
	GotoState('Idle');
}

//Items of the Affector chain join the Inventory chain, so, operating to one chain would mess up the other chain.
//It'll just remove the pickedup item from the Affector chain and will set this AI as the ONLY Affector.
//This AI will then call all the needed callbacks to the Affector in the Owner's Inventory chain. (The Affector is removed from the Affector chain, but not from the Inventory chain)
//NOTE: This leaves the bugs that come with UT stock Affector's code.
function SetOwnerWeapon() {
	local Inventory nextInv;
	
	if (bActive) {
		if (!Owner.IsA('Inventory')) {
			if (ownerWeapon != None)
				if (ownerWeapon.IsA('TournamentWeapon'))
					if (TournamentWeapon(ownerWeapon).Affector != Self) {
						lastAffector = TournamentWeapon(ownerWeapon).Affector;
						TournamentWeapon(ownerWeapon).Affector = None;
					}
		
			if (Owner.IsA('Pawn'))
				ownerWeapon = Pawn(Owner).Weapon;
			else
				for (nextInv = Owner.Inventory; nextInv != none; nextInv = nextInv.Inventory)
					if (nextInv.IsA('Weapon'))
						if (!Weapon(nextInv).bChangeWeapon) {
							ownerWeapon = Weapon(nextInv);
					
							break;
						}
							
			if (ownerWeapon != None)
				if (ownerWeapon.IsA('TournamentWeapon'))
					TournamentWeapon(ownerWeapon).Affector = Self;
		}
	}
}

function ChangedWeapon() {
	local Weapon OldWeapon;
	
	if (bActive) {
		bChagingWeapon = true;
		
		OldWeapon = ownerWeapon;
		SetOwnerWeapon();
		
		AMM.BaseAM.ChangedWeapon(Owner, OldWeapon, ownerWeapon);
		
		bChagingWeapon = false;
	}
}

auto state Pickup {
Begin:
	GotoState('Idle');
}

state Idle {

}

state Activated {
	function Activate() {
		bActive = true;
	}

	function EndState() {
		bActive = false;
		GotoState('Idle2');
	}
	
	function BeginState() {
		bActive = true;
	}
	
	function FireEffect() {
		AMM.BaseAM.PreFire(Owner, ownerWeapon, AMM.BaseAM.static.getAimedActor(Pawn(Owner), ownerWeapon));
		
		if (lastAffector != None)
			lastAffector.FireEffect();
	}
	
	function Trigger(Actor Other, Pawn EventInstigator) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other))
			AMM.BaseAM.Triggered(Owner, Other, EventInstigator);
	}
	
	function UnTrigger(Actor Other, Pawn EventInstigator) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other))
			AMM.BaseAM.UnTriggered(Owner, Other, EventInstigator);
	}
	
	function ZoneChange( ZoneInfo newZone ) {
		super.ZoneChange(newZone);
	
		AMM.BaseAM.ZoneChanged(Owner, ownerZone, newZone);
		ownerZone = newZone;
	}
	
	function HitWall( vector HitNormal, actor HitWall ) {
		AMM.BaseAM.HittingWall(Owner, HitNormal, HitWall);
	}
	
	function Landed( vector HitNormal ) {
		AMM.BaseAM.Landing(Owner, HitNormal);
	}
	
	
	function Touch( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other))
			AMM.BaseAM.Touching(Owner, Other);
	}
	
	function PostTouch( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other))
			AMM.BaseAM.Touched(Owner, Other);
	}
	
	function UnTouch( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other))
			AMM.BaseAM.UnTouched(Owner, Other);
	}
	
	//bBlockActors or bBlockPlayers on true, would cause a loop since the AI would encroach/bump with its owner. The AI will call them to the bumped/encroached actor,
	//feigning it really happened.
	
	function Bump( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other)) {
			Owner.Bump(Other);
			AMM.BaseAM.Bumped(Owner, Other);
		}
	}
	
	function bool EncroachingOn( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other)) {
			Owner.EncroachingOn(Other);
			AMM.BaseAM.EncroachedOn(Owner, Other);
			
			return true;
		}
		
		return false;
	}
	
	function EncroachedBy( Actor Other ) {
		if ((Other != Owner) && AMM.BaseAM.CanHaveAI(Other)) {
			Owner.EncroachedBy(Other);
			AMM.BaseAM.EncroachingBy(Owner, Other);
		}
	}
	
	function BaseChange() {
		AMM.BaseAM.ChangedBase(Owner);
	}

	function FellOutOfWorld() {
		AMM.BaseAM.FallingOutOfWorld(Owner);
	}

	function Tick( float Delta ) {
		local Actor AimedActor;
		local vector HitLocation, HitNormal, StartTrace, EndTrace;
	
		//The functions o() and w() are used to synchronize.
	
		if (!o()) {
			if (noneTimes >= AMM.BaseAM.maxNoneTimes) {
				if (!bNoneNotified) {
					AMM.BaseAM.BecameNone(Self);
					bNoneNotified = true;
				}
				//Don't let the mutator to have a None as lastAI.
				if (AMM.BaseAM.AMM.lastAI != Self)
					Destroy();
			}
			noneTimes++;
		} else {
			noneTimes = 0;
			bNoneNotified = false;
	
			if (AIP == None)
				GiveAIP();
	
			if (o()) Tag = Owner.Tag; //Receive Trigger functions.
			if (o()) SetLocation(Owner.Location);
			if (o()) SetRotation(Owner.Rotation);
			if (o()) if ((Owner.bCollideActors != bCollideActors) || (Owner.bBlockActors != bBlockActors) || (Owner.bBlockPlayers != bBlockPlayers))
				if (o()) SetCollision(Owner.bCollideActors, ((Owner.bBlockActors) && AMM.BaseAM.CanBlockActors(Owner)), (Owner.bBlockPlayers) && AMM.BaseAM.CanBlockPlayers(Owner));
			if (o()) bCollideWorld = Owner.bCollideWorld;
			if (o()) SetCollisionSize(Owner.CollisionRadius, Owner.CollisionHeight);
			if (o()) SetPhysics(Owner.Physics);
			
			if (o()) if (ownerState != String(Owner.getStateName())) {
				if (o()) AMM.BaseAM.StateChanged(Owner, ownerState, String(Owner.getStateName()));
				if (o()) ownerState = String(Owner.getStateName());
			}
			
			if (!bChagingWeapon)
				SetOwnerWeapon();
			
			if (o()) if (Owner.IsA('Pawn')) {
				if (o()) Pawn(Owner).bCountJumps = true;
				
				if (o() && w()) AimedActor = AMM.BaseAM.static.getAimedActor(Pawn(Owner), ownerWeapon, HitLocation, HitNormal, StartTrace, EndTrace);
				if (o() && w() && (AimedActor != none)) AMM.BaseAM.ActorAiming(Pawn(Owner), ownerWeapon, oldAimedActor, AimedActor, oldHitLocation, oldHitNormal, oldStartTrace, oldEndTrace, HitLocation, HitNormal, StartTrace, EndTrace);
				oldAimedActor = AimedActor;
				oldHitLocation = HitLocation;
				oldHitNormal = HitNormal;
				oldStartTrace = StartTrace;
				oldEndTrace = EndTrace;
			}
		}
		
		if (o()) checkValues();
		
		AMM.BaseAM.TickCallback(self, Owner);
	}
}

function bool o() {
	return AMM.BaseAM.static.exists(Owner);
}

function bool w() {
	return AMM.BaseAM.static.exists(ownerWeapon);
}