//=============================================================================
// ActorInfoPickup.
// 
// Date: 2016
// ©copyright Francesco Biscazzo. All rights reserved.
//=============================================================================
class ActorInfoPickup extends TournamentPickup;

var ActorInfo AI; //The ActorInfo this AIP is registered to.
var AdvancedMutatorManager AMM;

function RegisterActorInfo(ActorInfo AInfo, AdvancedMutatorManager AdvMutManager) {
	AI = AInfo;
	AMM = AdvMutManager;
	Activate();
}

function Destroyed() {
	//Remove this AIP from the Inventory chain of the Owner to not to mess its Inventory chain.
	removeFromInvChain();
	super.Destroyed();
	//Set this AIP so that this destruction doesn't need to synchronize with the Tick() function of the AI.
	AI.AIP = None;
}

function bool removeFromInvChain() {
	local Inventory prevInv, nextInv;
	
	if ((Owner == None) || (Owner.Inventory == None))
		return false;
	
	if (Owner.Inventory == Self) {
		Owner.Inventory = Inventory;
		
		return true;
	} else {
		prevInv = Owner.Inventory;
		for(nextInv = prevInv.Inventory; nextInv != None; nextInv = nextInv.Inventory) {
			if (nextInv == Self) {
				//Remove from Inv chain.
				prevInv.Inventory = Inventory;
						
				return true;
			}
			
			prevInv = nextInv;
		}
	}
			
	return false;
}

function addToChainFront() {
	//If already in the chain, move it to chain front.
	removeFromInvChain();
	Inventory = Owner.Inventory;
	Owner.Inventory = Self;
}

function Activate() {
	GotoState('Activated');
	Activate();
}

function Deactivate() {
	GotoState('Idle');
}

auto state Pickup {
Begin:
	GotoState('Idle');
}

state Idle {

}

state Activated
{
	function Activate() {
		bActive = true;
	}

	function EndState() {
		bActive = false;
		GotoState('Idle2');
	}
	
	function BeginState() {
		bActive = true;
	}
	
	function bool HandlePickupQuery(Inventory Item) {
		local bool bResult;
		
		bResult = Inventory.HandlePickupQuery(Item);
		if (!bResult)
			AMM.BaseAM.PickedUp(Owner, Item);
			
		return bResult;
	}
	
	function OwnerJumped() {
		super.OwnerJumped();
		
		AMM.BaseAM.Jumped(Owner);
	}

	function ChangedWeapon() {
		local Weapon OldWeapon;
		
		super.ChangedWeapon();
	
		AI.bChagingWeapon = true;
			
		OldWeapon = AI.ownerWeapon;
		AI.SetOwnerWeapon();
		
		AMM.BaseAM.ChangedWeapon(Owner, OldWeapon, AI.ownerWeapon);
		
		AI.bChagingWeapon = false;
	}
	
	function Weapon RecommendWeapon(out float rating, out int bUseAltMode) {
		local Weapon w;
		
		w = super.RecommendWeapon(rating, bUseAltMode);
		AMM.BaseAM.RecommendingWeapon(Owner, w, rating, bUseAltMode);
		
		return w;
	}
	
	function Inventory SelectNext() {
		local Inventory nextInv;
		
		nextInv = super.SelectNext();
		AMM.BaseAM.SelectingNext(Owner, nextInv);
		
		return nextInv;
	}
	
	function int ReduceDamage(int Damage, name DamageType, vector HitLocation) {
		local int ReducedAmount;
		
		if (Damage <= 0) {
			if (DamageType == 'Drowned')
				AMM.BaseAM.HeadInWater(Owner);
			else if (DamageType == 'Breathe')
				AMM.BaseAM.HeadOutOfWater(Owner);
		}
		
		ReducedAmount = super.ReduceDamage(Damage, DamageType, HitLocation);
		if (AMM.DamageAM != None)
			AMM.DamageAM.DamageReduced(Owner, Damage, ReducedAmount, DamageType, HitLocation);
		
		return ReducedAmount;
	}
	
	function Inventory PrioritizeArmor(int Damage, name DamageType, vector HitLocation) {
		local Inventory armor;
		
		armor = super.PrioritizeArmor(Damage, DamageType, HitLocation);
		if (AMM.DamageAM != None)
			AMM.DamageAM.PrioritizingArmor(Owner, armor, Damage, DamageType, HitLocation);
			
		return armor;
	}
	
	function SetOwnerDisplay() {
		
		super.SetOwnerDisplay();
		AMM.BaseAM.SettingDisplay(Owner);
	}
	
	function Tick(float Delta) {
		//Check if there are possibilities this AIP isn't in the Owner's Inventory chain.
		if ((Owner != None) && (Owner.Inventory != Self))
			addToChainFront();
	}
}